BOLTZMANN_CONSTANT = 1.380649e-23       # joules/Kelvin (kg m2 s-2 K-1)
AVOGADRO_CONSTANT = 6.02214076e23       # mol-1
GAS_CONSTANT = BOLTZMANN_CONSTANT * AVOGADRO_CONSTANT   # joules / Kelvin-mole (kg m2 s-2 K-1 mol-1)
